"""
Classes and utility functions for GSWorkspace and GSStepCache.
"""
import atexit
import datetime
import json
import logging
import os
import time
from dataclasses import dataclass
from pathlib import Path
from typing import List, Optional, Union

import google.auth
from google.api_core import exceptions
from google.auth.credentials import Credentials
from google.cloud import storage
from google.oauth2.credentials import Credentials as OAuth2Credentials
from google.oauth2.service_account import Credentials as ServiceAccountCredentials

from tango.common.aliases import PathOrStr
from tango.common.exceptions import TangoError
from tango.common.remote_utils import RemoteConstants
from tango.step import Step
from tango.step_info import StepInfo

logger = logging.getLogger(__name__)


def empty_bucket(bucket_name: str):
    """
    Removes all the tango-related blobs from the specified bucket.
    Used for testing.
    """
    credentials, project = google.auth.default()
    client = storage.Client(project=project, credentials=credentials)
    bucket = client.bucket(bucket_name)
    try:
        bucket.delete_blobs(list(bucket.list_blobs(prefix="tango-")))
    except exceptions.NotFound:
        pass


def empty_datastore(namespace: str):
    """
    Removes all the tango-related entities from the specified namespace in datastore.
    Used for testing.
    """
    from google.cloud import datastore

    credentials, project = google.auth.default()
    client = datastore.Client(project=project, credentials=credentials, namespace=namespace)
    run_query = client.query(kind="run")
    run_query.keys_only()
    keys = [entity.key for entity in run_query.fetch()]
    stepinfo_query = client.query(kind="stepinfo")
    stepinfo_query.keys_only()
    keys += [entity.key for entity in stepinfo_query.fetch()]
    client.delete_multi(keys)


@dataclass
class GSArtifact:
    """
    A GSArtifact object is used for representing storage objects in google cloud storage.
    """

    name: str
    """
    Name of the artifact.
    """
    artifact_path: str
    """
    Remote location url for the artifact.
    """
    created: datetime.datetime
    """
    Time of creation.
    """
    committed: bool
    """
    If set to True, no further changes to the remote artifact are allowed.
    If set to False, it means that the artifact is under construction.
    """


class GSArtifactConflict(TangoError):
    """
    Error denoting that the storage artifact already exists.
    """

    pass


class GSArtifactNotFound(TangoError):
    """
    Error denoting that the storage artifact does not exist.
    """

    pass


class GSArtifactWriteError(TangoError):
    """
    Error denoting that there was an issue writing the artifact to the remote storage.
    """

    pass


class GSClient:
    """
    A client for interacting with Google Cloud Storage. The authorization works by
    providing OAuth2 credentials.

    :param bucket_name: The name of the Google Cloud bucket to use.
    :param credentials: OAuth2 credentials can be provided. If not provided, default
        gcloud credentials are inferred.
    :param project: Optionally, the project ID can be provided. This is not essential
        for `google.cloud.storage` API, since buckets are at the account level, rather
        than the project level.
    """

    placeholder_file = ".placeholder"
    """
    The placeholder file is used for creation of a folder in the cloud bucket,
    as empty folders are not allowed. It is also used as a marker for the creation
    time of the folder, hence we use a separate file to mark the artifact as
    uncommitted.
    """

    uncommitted_file = ".uncommitted"
    """
    The uncommitted file is used to denote an artifact under construction.
    """

    settings_file = "settings.json"
    """
    This file is for storing metadata like version information, etc.
    """

    NUM_CONCURRENT_WORKERS: int = 9

    def __init__(
        self,
        bucket_name: str,
        credentials: Optional[Credentials] = None,
        project: Optional[str] = None,
    ):
        if not credentials:
            credentials, project = google.auth.default()

        self.storage = storage.Client(project=project, credentials=credentials)
        self.bucket_name = bucket_name

        blob = self.storage.bucket(bucket_name).blob(self.settings_file)  # no HTTP request yet
        try:
            with blob.open("r") as file_ref:
                json.load(file_ref)
        except exceptions.NotFound:
            settings = {"version": 1}
            with blob.open("w") as file_ref:
                json.dump(settings, file_ref)

    def url(self, artifact: Optional[str] = None):
        """
        Returns the remote url of the storage artifact.
        """
        path = f"gs://{self.bucket_name}"
        if artifact is not None:
            path = f"{path}/{artifact}"
        return path

    @classmethod
    def _convert_blobs_to_artifact(cls, blobs: List[storage.Blob]) -> GSArtifact:
        """
        Converts a list of `google.cloud.storage.Blob` to a `GSArtifact`.
        """
        name: str
        artifact_path: str
        created: datetime.datetime
        committed: bool = True

        for blob in blobs:
            if blob.name.endswith(cls.placeholder_file):
                created = blob.time_created
                name = blob.name.replace("/" + cls.placeholder_file, "")
                artifact_path = name  # does not contain bucket info here.
            elif blob.name.endswith(cls.uncommitted_file):
                committed = False

        assert name is not None, "Folder is not a GSArtifact, should not have happened."
        return GSArtifact(name, artifact_path, created, committed)

    @classmethod
    def from_env(cls, bucket_name: str):
        """
        Constructs the client object from the environment, using default credentials.
        """
        return cls(bucket_name)

    def get(self, artifact: Union[str, GSArtifact]) -> GSArtifact:
        """
        Returns a `GSArtifact` object created by fetching the artifact's information
        from remote location.
        """
        if isinstance(artifact, str):
            path = artifact
        else:
            # We have an artifact, and we recreate it with refreshed info.
            path = artifact.artifact_path

        blobs = list(self.storage.bucket(self.bucket_name).list_blobs(prefix=path))
        if len(blobs) > 0:
            return self._convert_blobs_to_artifact(blobs)
        else:
            raise GSArtifactNotFound()

    @classmethod
    def _gs_path(cls, *args):
        """
        Returns path within google cloud storage bucket. We use this since we cannot
        use `os.path.join` for cloud storage paths.
        """
        return "/".join(args)

    def create(self, artifact: str):
        """
        Creates a new artifact in the remote location. By default, it is uncommitted.
        """
        bucket = self.storage.bucket(self.bucket_name)
        # gives refreshed information
        if bucket.blob(self._gs_path(artifact, self.placeholder_file)).exists():
            raise GSArtifactConflict(f"{artifact} already exists!")
        else:
            # Additional safety check
            if bucket.blob(self._gs_path(artifact, self.uncommitted_file)).exists():
                raise GSArtifactConflict(f"{artifact} already exists!")
            bucket.blob(self._gs_path(artifact, self.placeholder_file)).upload_from_string("")
            bucket.blob(self._gs_path(artifact, self.uncommitted_file)).upload_from_string("")
        return self._convert_blobs_to_artifact(list(bucket.list_blobs(prefix=artifact)))

    def delete(self, artifact: GSArtifact):
        """
        Removes the artifact from the remote location.
        """
        bucket = self.storage.bucket(self.bucket_name)
        blobs = list(bucket.list_blobs(prefix=artifact.artifact_path))
        bucket.delete_blobs(blobs)

    def upload(self, artifact: Union[str, GSArtifact], objects_dir: Path):
        """
        Writes the contents of objects_dir to the remote artifact location.
        """
        if isinstance(artifact, str):
            folder_path = artifact
        else:
            folder_path = artifact.artifact_path

        source_path = str(objects_dir)

        def _sync_blob(source_file_path: str, target_file_path: str):
            blob = self.storage.bucket(self.bucket_name).blob(target_file_path)
            blob.upload_from_filename(source_file_path)

        import concurrent.futures

        try:
            # TODO: google-cloud-storage==2.7.0 has added a preview feature called `transfer_manager`
            # which allows for concurrent uploads and downloads. We should upgrade to this once
            # it is more robustly supported. Also update in `download()`.
            if os.path.isdir(source_path):
                with concurrent.futures.ThreadPoolExecutor(
                    max_workers=self.NUM_CONCURRENT_WORKERS, thread_name_prefix="GSClient.upload()-"
                ) as executor:
                    upload_futures = []
                    for dirpath, _, filenames in os.walk(source_path):
                        for filename in filenames:
                            source_file_path = os.path.join(dirpath, filename)
                            target_file_path = self._gs_path(
                                folder_path, source_file_path.replace(source_path + "/", "")
                            )
                        upload_futures.append(
                            executor.submit(_sync_blob, source_file_path, target_file_path)
                        )
                    for future in concurrent.futures.as_completed(upload_futures):
                        future.result()
            else:
                source_file_path = source_path
                target_file_path = self._gs_path(folder_path, os.path.basename(source_file_path))
                _sync_blob(source_file_path, target_file_path)
        except Exception:
            raise GSArtifactWriteError()

    def commit(self, artifact: Union[str, GSArtifact]):
        """
        Marks the artifact as committed. No further changes to the artifact are allowed.
        """
        if isinstance(artifact, str):
            folder_path = artifact
        else:
            folder_path = artifact.artifact_path
        bucket = self.storage.bucket(self.bucket_name)
        try:
            bucket.delete_blob(self._gs_path(folder_path, self.uncommitted_file))
        except exceptions.NotFound:
            if not bucket.blob(self._gs_path(folder_path, self.placeholder_file)).exists():
                raise GSArtifactNotFound()
            # Otherwise, already committed. No change.

    def download(self, artifact: GSArtifact, target_dir: PathOrStr):
        """
        Writes the contents of the remote artifact to the `target_dir`.
        """
        assert (
            self.storage.bucket(self.bucket_name)
            .blob(self._gs_path(artifact.artifact_path, self.placeholder_file))
            .exists()
        )

        def _fetch_blob(blob: storage.Blob):
            source_path = blob.name.replace(artifact.artifact_path + "/", "")
            target_path = os.path.join(target_dir, source_path)
            if not os.path.exists(os.path.dirname(target_path)):
                os.mkdir(os.path.dirname(target_path))
            blob.download_to_filename(target_path)

        import concurrent.futures

        bucket = self.storage.bucket(self.bucket_name)
        bucket.update()

        try:
            with concurrent.futures.ThreadPoolExecutor(
                max_workers=self.NUM_CONCURRENT_WORKERS, thread_name_prefix="GSClient.download()-"
            ) as executor:
                download_futures = []
                for blob in bucket.list_blobs(self.bucket_name, prefix=artifact.artifact_path):
                    download_futures.append(executor.submit(_fetch_blob, blob))
                for future in concurrent.futures.as_completed(download_futures):
                    future.result()
        except exceptions.NotFound:
            raise GSArtifactWriteError()

    def artifacts(self, prefix: str, uncommitted: bool = True) -> List[GSArtifact]:
        """
        Lists all the artifacts within the remote storage, based on
        `match` and `uncommitted` criteria. These can include steps and runs.
        """
        list_of_artifacts = []
        for folder_name in self.storage.list_blobs(
            self.bucket_name, prefix=prefix, delimiter="/"
        )._get_next_page_response()["prefixes"]:
            artifact = self._convert_blobs_to_artifact(
                list(self.storage.list_blobs(self.bucket_name, prefix=folder_name))
            )
            if not uncommitted:
                if not artifact.committed:
                    continue
            list_of_artifacts.append(artifact)
        return list_of_artifacts


def get_credentials(credentials: Optional[Union[str, Credentials]] = None) -> Credentials:
    """
    :param credentials:
        * if OAuth2 credentials are provided, they are returned.
        * if `str`, it can be either a file path or a json string of credentials dict.
        * if `None`, credentials are inferred from the environment.

    More details on Google Cloud credentials can be found here:
    https://googleapis.dev/python/google-auth/latest/user-guide.html#service-account-private-key-files,
    and https://googleapis.dev/python/google-api-core/latest/auth.html
    """

    # BeakerExecutor uses GOOGLE_TOKEN
    credentials = os.environ.get("GOOGLE_TOKEN", credentials)
    if credentials is not None:
        # Path to the credentials file has been provided
        if isinstance(credentials, str) and credentials.endswith(".json"):
            with open(credentials) as file_ref:
                credentials = file_ref.read()
        try:
            # If credentials dict has been passed as a json string
            credentials_dict = json.loads(credentials)
            if credentials_dict.pop("type", None) == "service_account":
                credentials = ServiceAccountCredentials.from_service_account_info(credentials_dict)
            else:
                # sometimes the credentials dict may not contain `token` and `token_uri` keys,
                # but `Credentials()` needs the parameter.
                token = credentials_dict.pop("token", None)
                token_uri = credentials_dict.pop("token_uri", "https://oauth2.googleapis.com/token")
                credentials = OAuth2Credentials(
                    token=token, token_uri=token_uri, **credentials_dict
                )
        except (json.decoder.JSONDecodeError, TypeError, ValueError):
            # It is not a json string.
            # We use this string because BeakerExecutor cannot write a None secret.
            if credentials == "default":
                credentials = None
    if not credentials:
        # Infer default credentials
        credentials, _ = google.auth.default()
    return credentials


def get_client(
    bucket_name: str,
    credentials: Optional[Union[str, Credentials]] = None,
    project: Optional[str] = None,
) -> GSClient:
    """
    Returns a `GSClient` object for a google cloud bucket.
    """
    credentials = get_credentials(credentials)
    return GSClient(bucket_name, credentials=credentials, project=project)


class Constants(RemoteConstants):
    pass


class GCSStepLock:
    """
    Google Cloud offers consistency https://cloud.google.com/storage/docs/consistency,
    so we can use lock files.
    """

    def __init__(
        self,
        client: GSClient,
        step: Union[str, StepInfo, Step],
    ):
        self._client = client
        self._step_id = step if isinstance(step, str) else step.unique_id
        self._lock_artifact_name = RemoteConstants.step_lock_artifact_name(step)
        self._lock_artifact: Optional[GSArtifact] = None
        self.lock_artifact_url = self._client.url(self._lock_artifact_name)

    def acquire(self, timeout=None, poll_interval: float = 2.0, log_interval: float = 30.0) -> None:
        if self._lock_artifact is not None:
            return
        start = time.monotonic()
        last_logged = None
        while timeout is None or (time.monotonic() - start < timeout):
            try:
                self._lock_artifact = self._client.create(self._lock_artifact_name)
                atexit.register(self.release)

            except GSArtifactConflict:
                if last_logged is None or last_logged - start >= log_interval:
                    logger.warning(
                        "Waiting to acquire lock artifact for step '%s':\n\n%s\n\n"
                        "This probably means the step is being run elsewhere, but if you're sure it isn't "
                        "you can just delete the lock artifact, using the command: \n`gsutil rm -r %s`",
                        self._step_id,
                        self.lock_artifact_url,
                        self.lock_artifact_url,
                    )
                    last_logged = time.monotonic()
                time.sleep(poll_interval)
                continue
            else:
                break
        else:
            raise TimeoutError(
                f"Timeout error occurred while waiting to acquire artifact lock for step '{self._step_id}':\n\n"
                f"{self.lock_artifact_url}\n\n"
                f"This probably means the step is being run elsewhere, but if you're sure it isn't you can "
                f"just delete the lock, using the command: \n`gsutil rm -r {self.lock_artifact_url}`"
            )

    def release(self):
        if self._lock_artifact is not None:
            try:
                self._client.delete(self._lock_artifact)
            except GSArtifactNotFound:
                # Artifact must have been manually deleted.
                pass
            self._lock_artifact = None
            atexit.unregister(self.release)

    def __del__(self):
        self.release()



import os
import subprocess
from typing import List, Optional, Union

from tango.common import PathOrStr, RegistrableFunction, make_registrable
from tango.step import Step


@make_registrable(exist_ok=True)
def check_path_existence(path: PathOrStr):
    assert os.path.exists(path), f"Output not found at {path}!"


@Step.register("shell_step")
class ShellStep(Step):
    """
    This step runs a shell command, and returns the standard output as a string.

    .. tip::

        Registered as a :class:`~tango.step.Step` under the name "shell_step".

    :param shell_command: The shell command to run.
    :param output_path: The step makes no assumptions about the command being run. If your command produces some
        output, you can optionally specify the output path for recording the output location, and optionally
        validating it. See `validate_output` argument for this.
    :param validate_output: If an expected `output_path` has been specified, you can choose to validate that the
        step produced the correct output. By default, it will just check if the `output_path` exists, but you can
        pass any other validating function. For example, if your command is a script generating a model output,
        you can check if the model weights can be loaded.
    :param kwargs: Other kwargs to be passed to `subprocess.run()`. If you need to take advantage of environment
        variables, set `shell = True`.
    """

    def run(  # type: ignore[override]
        self,
        shell_command: Union[str, List[str]],
        output_path: Optional[PathOrStr] = None,
        validate_output: RegistrableFunction = check_path_existence,
        **kwargs,
    ):
        output = self.run_command(shell_command, **kwargs)
        self.logger.info(output)
        if output_path is not None:
            validate_output(output_path)
            self.logger.info(f"Output found at: {output_path}")

        return str(output.decode("utf-8"))

    def run_command(self, command: Union[str, List[str]], **kwargs):
        import shlex

        if kwargs.get("shell", None):
            if isinstance(command, list):
                command = shlex.join(command)
        else:
            if isinstance(command, str):
                command = shlex.split(command)
        self.logger.info(f"Command: {command}")
        process = subprocess.run(command, capture_output=True, **kwargs)
        if process.returncode != 0:
            raise RuntimeError(f"The command failed with the following errors: {process.stderr}")
        return process.stdout



import logging
from typing import Union

from tango.step import Step
from tango.step_info import StepInfo

logger = logging.getLogger(__name__)


class RemoteConstants:
    """
    Common constants to be used as prefixes and filenames in remote workspaces.
    """

    RUN_ARTIFACT_PREFIX: str = "tango-run-"
    RUN_DATA_FNAME: str = "run.json"
    STEP_ARTIFACT_PREFIX: str = "tango-step-"
    STEP_INFO_FNAME: str = "step_info.json"
    STEP_RESULT_DIR: str = "result"
    STEP_GRAPH_ARTIFACT_PREFIX: str = "tango-step-graph-"
    STEP_EXPERIMENT_PREFIX: str = "tango-step-"
    STEP_GRAPH_FILENAME: str = "config.json"
    GITHUB_TOKEN_SECRET_NAME: str = "TANGO_GITHUB_TOKEN"
    RESULTS_DIR: str = "/tango/output"
    INPUT_DIR: str = "/tango/input"
    LOCK_ARTIFACT_SUFFIX: str = "-lock"

    @classmethod
    def step_artifact_name(cls, step: Union[str, StepInfo, Step]) -> str:
        return f"{cls.STEP_ARTIFACT_PREFIX}{step if isinstance(step, str) else step.unique_id}"

    @classmethod
    def step_lock_artifact_name(cls, step: Union[str, StepInfo, Step]) -> str:
        return f"{cls.step_artifact_name(step)}{cls.LOCK_ARTIFACT_SUFFIX}"

    @classmethod
    def run_artifact_name(cls, name: str) -> str:
        return f"{cls.RUN_ARTIFACT_PREFIX}{name}"



import json
import logging
import os
import random
from collections import OrderedDict
from pathlib import Path
from typing import Dict, List, Optional, Type, TypeVar, Union, cast
from urllib.parse import ParseResult

import petname
from beaker import Dataset
from beaker import Dataset as BeakerDataset
from beaker import (
    DatasetConflict,
    DatasetNotFound,
    DatasetSort,
    Digest,
    Experiment,
    ExperimentNotFound,
)

from tango.common.util import make_safe_filename, tango_cache_dir
from tango.step import Step
from tango.step_info import StepInfo, StepState
from tango.workspace import Run, RunInfo, RunSort, StepInfoSort, Workspace
from tango.workspaces.remote_workspace import RemoteWorkspace

from .common import BeakerStepLock, Constants, dataset_url, get_client
from .step_cache import BeakerStepCache

T = TypeVar("T")
U = TypeVar("U", Run, StepInfo)

logger = logging.getLogger(__name__)


@Workspace.register("beaker")
class BeakerWorkspace(RemoteWorkspace):
    """
    This is a :class:`~tango.workspace.Workspace` that stores step artifacts on `Beaker`_.

    .. tip::
        Registered as a :class:`~tango.workspace.Workspace` under the name "beaker".

    :param workspace: The name or ID of the Beaker workspace to use.
    :param kwargs: Additional keyword arguments passed to :meth:`Beaker.from_env() <beaker.Beaker.from_env()>`.
    """

    STEP_INFO_CACHE_SIZE = 512
    Constants = Constants
    NUM_CONCURRENT_WORKERS = 9

    def __init__(self, workspace: str, max_workers: Optional[int] = None, **kwargs):
        self.beaker = get_client(beaker_workspace=workspace, **kwargs)
        self._cache = BeakerStepCache(beaker=self.beaker)
        self._locks: Dict[Step, BeakerStepLock] = {}
        super().__init__()
        self.max_workers = max_workers
        self._disk_cache_dir = tango_cache_dir() / "beaker_cache" / "_objects"
        self._mem_cache: "OrderedDict[Digest, Union[StepInfo, Run]]" = OrderedDict()

    @property
    def cache(self):
        return self._cache

    @property
    def locks(self):
        return self._locks

    @property
    def steps_dir_name(self):
        return "beaker_workspace"

    @property
    def url(self) -> str:
        return f"beaker://{self.beaker.workspace.get().full_name}"

    def _step_location(self, step: Step) -> str:
        return dataset_url(self.beaker, self.Constants.step_artifact_name(step))

    @classmethod
    def from_parsed_url(cls, parsed_url: ParseResult) -> Workspace:
        workspace: str
        if parsed_url.netloc and parsed_url.path:
            # e.g. "beaker://ai2/my-workspace"
            workspace = parsed_url.netloc + parsed_url.path
        elif parsed_url.netloc:
            # e.g. "beaker://my-workspace"
            workspace = parsed_url.netloc
        else:
            raise ValueError(f"Bad URL for Beaker workspace '{parsed_url}'")
        return cls(workspace)

    @property
    def current_beaker_experiment(self) -> Optional[Experiment]:
        """
        When the workspace is being used within a Beaker experiment that was submitted
        by the Beaker executor, this will return the `Experiment` object.
        """
        experiment_name = os.environ.get("BEAKER_EXPERIMENT_NAME")
        if experiment_name is not None:
            try:
                return self.beaker.experiment.get(experiment_name)
            except ExperimentNotFound:
                return None
        else:
            return None

    def _remote_lock(self, step: Step) -> BeakerStepLock:
        return BeakerStepLock(
            self.beaker, step, current_beaker_experiment=self.current_beaker_experiment
        )

    def _get_object_from_cache(self, digest: Digest, o_type: Type[U]) -> Optional[U]:
        cache_path = self._disk_cache_dir / make_safe_filename(str(digest))
        if digest in self._mem_cache:
            cached = self._mem_cache.pop(digest)
            # Move to end.
            self._mem_cache[digest] = cached
            return cached if isinstance(cached, o_type) else None
        elif cache_path.is_file():
            try:
                with cache_path.open("r+t") as f:
                    json_dict = json.load(f)
                    cached = o_type.from_json_dict(json_dict)
            except Exception as exc:
                logger.warning("Error while loading object from workspace cache: %s", str(exc))
                try:
                    os.remove(cache_path)
                except FileNotFoundError:
                    pass
                return None
            # Add to in-memory cache.
            self._mem_cache[digest] = cached
            while len(self._mem_cache) > self.STEP_INFO_CACHE_SIZE:
                self._mem_cache.popitem(last=False)
            return cached  # type: ignore
        else:
            return None

    def _add_object_to_cache(self, digest: Digest, o: U):
        self._disk_cache_dir.mkdir(parents=True, exist_ok=True)
        cache_path = self._disk_cache_dir / make_safe_filename(str(digest))
        self._mem_cache[digest] = o
        with cache_path.open("w+t") as f:
            json.dump(o.to_json_dict(), f)
        while len(self._mem_cache) > self.STEP_INFO_CACHE_SIZE:
            self._mem_cache.popitem(last=False)

    def step_info(self, step_or_unique_id: Union[Step, str]) -> StepInfo:
        try:
            dataset = self.beaker.dataset.get(self.Constants.step_artifact_name(step_or_unique_id))
            return self._get_step_info_from_dataset(dataset)
        except (DatasetNotFound, FileNotFoundError):
            if not isinstance(step_or_unique_id, Step):
                raise KeyError(step_or_unique_id)
            step_info = StepInfo.new_from_step(step_or_unique_id)
            self._update_step_info(step_info)
            return step_info

    def _get_step_info_from_dataset(self, dataset: Dataset) -> StepInfo:
        file_info = self.beaker.dataset.file_info(dataset, Constants.STEP_INFO_FNAME)
        step_info: StepInfo
        cached = (
            None
            if file_info.digest is None
            else self._get_object_from_cache(file_info.digest, StepInfo)
        )
        if cached is not None:
            step_info = cached
        else:
            step_info_bytes = self.beaker.dataset.get_file(dataset, file_info, quiet=True)
            step_info = StepInfo.from_json_dict(json.loads(step_info_bytes))
            if file_info.digest is not None:
                self._add_object_to_cache(file_info.digest, step_info)
        return step_info

    def _save_run(
        self, steps: Dict[str, StepInfo], run_data: Dict[str, str], name: Optional[str] = None
    ) -> Run:
        # Create a remote dataset that represents this run. The dataset which just contain
        # a JSON file that maps step names to step unique IDs.
        run_dataset: BeakerDataset
        if name is None:
            # Find a unique name to use.
            while True:
                name = petname.generate() + str(random.randint(0, 100))
                try:
                    run_dataset = self.beaker.dataset.create(
                        self.Constants.run_artifact_name(cast(str, name)), commit=False
                    )
                except DatasetConflict:
                    continue
                else:
                    break
        else:
            try:
                run_dataset = self.beaker.dataset.create(
                    self.Constants.run_artifact_name(name), commit=False
                )
            except DatasetConflict:
                raise ValueError(f"Run name '{name}' is already in use")

        # Upload run data to dataset.
        # NOTE: We don't commit the dataset here since we'll need to upload the logs file
        # after the run.
        self.beaker.dataset.upload(
            run_dataset, json.dumps(run_data).encode(), self.Constants.RUN_DATA_FNAME, quiet=True
        )

        return Run(name=cast(str, name), steps=steps, start_date=run_dataset.created)

    def registered_runs(self) -> Dict[str, Run]:
        import concurrent.futures

        runs: Dict[str, Run] = {}

        with concurrent.futures.ThreadPoolExecutor(
            max_workers=self.NUM_CONCURRENT_WORKERS,
            thread_name_prefix="BeakerWorkspace.registered_runs()-",
        ) as executor:
            run_futures = []
            for dataset in self.beaker.workspace.iter_datasets(
                match=self.Constants.RUN_ARTIFACT_PREFIX, uncommitted=True, results=False
            ):
                run_futures.append(executor.submit(self._get_run_from_dataset, dataset))
            for future in concurrent.futures.as_completed(run_futures):
                run = future.result()
                if run is not None:
                    runs[run.name] = run

        return runs

    def search_registered_runs(
        self,
        *,
        sort_by: Optional[RunSort] = None,
        sort_descending: bool = True,
        match: Optional[str] = None,
        start: Optional[int] = None,
        stop: Optional[int] = None,
    ) -> List[RunInfo]:
        if match is None:
            match = Constants.RUN_ARTIFACT_PREFIX
        else:
            match = Constants.RUN_ARTIFACT_PREFIX + match

        if sort_by is None or sort_by == RunSort.START_DATE:
            sort = DatasetSort.created
        elif sort_by == RunSort.NAME:
            sort = DatasetSort.dataset_name
        else:
            raise NotImplementedError

        runs = []
        for dataset in self.beaker.workspace.iter_datasets(
            match=match,
            results=False,
            cursor=start or 0,
            limit=None if stop is None else stop - (start or 0),
            sort_by=sort,
            descending=sort_descending,
        ):
            if dataset.name is not None and dataset.name.startswith(
                self.Constants.RUN_ARTIFACT_PREFIX
            ):
                run_name = dataset.name[len(self.Constants.RUN_ARTIFACT_PREFIX) :]
                runs.append(RunInfo(name=run_name, start_date=dataset.created))

        return runs

    def num_registered_runs(self, *, match: Optional[str] = None) -> int:
        if match is None:
            match = Constants.RUN_ARTIFACT_PREFIX
        else:
            match = Constants.RUN_ARTIFACT_PREFIX + match

        count = 0
        for dataset in self.beaker.workspace.iter_datasets(
            match=match,
            results=False,
        ):
            if dataset.name is not None and dataset.name.startswith(Constants.RUN_ARTIFACT_PREFIX):
                count += 1

        return count

    def search_step_info(
        self,
        *,
        sort_by: Optional[StepInfoSort] = None,
        sort_descending: bool = True,
        match: Optional[str] = None,
        state: Optional[StepState] = None,
        start: int = 0,
        stop: Optional[int] = None,
    ) -> List[StepInfo]:
        if state is not None:
            raise NotImplementedError(
                f"{self.__class__.__name__} cannot filter steps efficiently by state"
            )

        if match is None:
            match = Constants.STEP_ARTIFACT_PREFIX
        else:
            match = Constants.STEP_ARTIFACT_PREFIX + match

        sort: Optional[DatasetSort] = None
        if sort_by is None or sort_by == StepInfoSort.START_TIME:
            sort = DatasetSort.created
        elif sort_by == StepInfoSort.UNIQUE_ID:
            sort = DatasetSort.dataset_name
        elif sort_by is not None:
            raise NotImplementedError

        steps = []
        for dataset in self.beaker.workspace.iter_datasets(
            match=match,
            results=False,
            cursor=start or 0,
            limit=None if stop is None else stop - (start or 0),
            sort_by=sort or DatasetSort.created,
            descending=sort_descending,
        ):
            try:
                steps.append(self._get_step_info_from_dataset(dataset))
            except (DatasetNotFound, FileNotFoundError):
                continue

        return steps

    def num_steps(self, *, match: Optional[str] = None, state: Optional[StepState] = None) -> int:
        if state is not None:
            raise NotImplementedError(
                f"{self.__class__.__name__} cannot filter steps efficiently by state"
            )

        if match is None:
            match = Constants.STEP_ARTIFACT_PREFIX
        else:
            match = Constants.STEP_ARTIFACT_PREFIX + match

        count = 0
        for dataset in self.beaker.workspace.iter_datasets(
            match=match,
            results=False,
        ):
            if dataset.name is not None and dataset.name.startswith(Constants.STEP_ARTIFACT_PREFIX):
                count += 1

        return count

    def registered_run(self, name: str) -> Run:
        err_msg = f"Run '{name}' not found in workspace"

        try:
            dataset_for_run = self.beaker.dataset.get(self.Constants.run_artifact_name(name))
            # Make sure the run is in our workspace.
            if dataset_for_run.workspace_ref.id != self.beaker.workspace.get().id:  # type: ignore # TODO
                raise DatasetNotFound
        except DatasetNotFound:
            raise KeyError(err_msg)

        run = self._get_run_from_dataset(dataset_for_run)
        if run is None:
            raise KeyError(err_msg)
        else:
            return run

    def _save_run_log(self, name: str, log_file: Path):
        run_dataset = self.Constants.run_artifact_name(name)
        self.beaker.dataset.sync(run_dataset, log_file, quiet=True)
        self.beaker.dataset.commit(run_dataset)

    def _get_run_from_dataset(self, dataset: BeakerDataset) -> Optional[Run]:
        if dataset.name is None:
            return None
        if not dataset.name.startswith(self.Constants.RUN_ARTIFACT_PREFIX):
            return None

        try:
            run_name = dataset.name[len(self.Constants.RUN_ARTIFACT_PREFIX) :]
            steps_info_bytes = self.beaker.dataset.get_file(
                dataset, self.Constants.RUN_DATA_FNAME, quiet=True
            )
            steps_info = json.loads(steps_info_bytes)
        except (DatasetNotFound, FileNotFoundError):
            return None

        steps: Dict[str, StepInfo] = {}
        import concurrent.futures

        with concurrent.futures.ThreadPoolExecutor(
            max_workers=self.NUM_CONCURRENT_WORKERS,
            thread_name_prefix="BeakerWorkspace._get_run_from_dataset()-",
        ) as executor:
            step_info_futures = []
            for unique_id in steps_info.values():
                step_info_futures.append(executor.submit(self.step_info, unique_id))
            for future in concurrent.futures.as_completed(step_info_futures):
                step_info = future.result()
                assert step_info.step_name is not None
                steps[step_info.step_name] = step_info

        return Run(name=run_name, start_date=dataset.created, steps=steps)

    def _update_step_info(self, step_info: StepInfo):
        dataset_name = self.Constants.step_artifact_name(step_info)

        step_info_dataset: BeakerDataset
        try:
            self.beaker.dataset.create(dataset_name, commit=False)
        except DatasetConflict:
            pass
        step_info_dataset = self.beaker.dataset.get(dataset_name)

        self.beaker.dataset.upload(
            step_info_dataset,  # folder name
            json.dumps(step_info.to_json_dict()).encode(),  # step info dict.
            self.Constants.STEP_INFO_FNAME,  # step info filename
            quiet=True,
        )



"""
.. important::
    To use this integration you should install ``tango`` with the "gs" extra
    (e.g. ``pip install tango[gs]``) or just install the `gcsfs <https://gcsfs.readthedocs.io/>`_
    library after the fact (e.g. ``pip install gcsfs``).

Components for Tango integration with `GS <https://cloud.google.com/storage/>`_.
"""

from tango.common.exceptions import IntegrationMissingError

try:
    from google.cloud import datastore, storage
except (ModuleNotFoundError, ImportError):
    raise IntegrationMissingError("gs", dependencies={"google-cloud-storage"})

from .step_cache import GSStepCache
from .workspace import GSWorkspace

__all__ = [
    "GSStepCache",
    "GSWorkspace",
]



import logging
from pathlib import Path
from typing import Optional, Union

from tango.common import PathOrStr
from tango.common.util import make_safe_filename, tango_cache_dir
from tango.integrations.gs.common import (
    Constants,
    GSArtifact,
    GSArtifactConflict,
    GSArtifactNotFound,
    GSArtifactWriteError,
    GSClient,
)
from tango.step import Step
from tango.step_cache import StepCache
from tango.step_caches.remote_step_cache import RemoteNotFoundError, RemoteStepCache
from tango.step_info import StepInfo

logger = logging.getLogger(__name__)


@StepCache.register("gs")
class GSStepCache(RemoteStepCache):
    """
    This is a :class:`~tango.step_cache.StepCache` that's used by :class:`GSWorkspace`.
    It stores the results of steps on Google cloud buckets as blobs.

    It also keeps a limited in-memory cache as well as a local backup on disk, so fetching a
    step's resulting subsequent times should be fast.

    .. tip::
        Registered as a :class:`~tango.step_cache.StepCache` under the name "gs".

    :param bucket_name: The name of the google cloud bucket to use.
    :param client: The google cloud storage client to use.
    """

    Constants = Constants

    def __init__(self, bucket_name: str, client: Optional[GSClient] = None):
        if client is not None:
            assert (
                bucket_name == client.bucket_name
            ), "Assert that bucket name is same as client bucket until we do better"
            self.bucket_name = bucket_name
            self._client = client
        else:
            self._client = GSClient(bucket_name)
        super().__init__(
            tango_cache_dir() / "gs_cache" / make_safe_filename(self._client.bucket_name)
        )

    @property
    def client(self):
        return self._client

    def _step_result_remote(self, step: Union[Step, StepInfo]) -> Optional[GSArtifact]:
        """
        Returns a `GSArtifact` object containing the details of the step.
        This only returns if the step has been finalized (committed).
        """
        try:
            artifact = self.client.get(self.Constants.step_artifact_name(step))
            return artifact if artifact.committed else None
        except GSArtifactNotFound:
            return None

    def _upload_step_remote(self, step: Step, objects_dir: Path) -> GSArtifact:
        """
        Uploads the step's output to remote location.
        """
        artifact_name = self.Constants.step_artifact_name(step)
        try:
            self.client.create(artifact_name)
        except GSArtifactConflict:
            pass
        try:
            self.client.upload(artifact_name, objects_dir)
            self.client.commit(artifact_name)
        except GSArtifactWriteError:
            pass

        return self.client.get(artifact_name)

    def _download_step_remote(self, step_result, target_dir: PathOrStr) -> None:
        """
        Downloads the step's output from remote location.
        """
        try:
            self.client.download(step_result, target_dir)
        except GSArtifactNotFound:
            raise RemoteNotFoundError()

    def __len__(self):
        """
        Returns the number of committed step outputs present in the remote location.
        """
        # NOTE: lock files should not count here.
        return sum(
            1
            for ds in self.client.artifacts(
                prefix=self.Constants.STEP_ARTIFACT_PREFIX, uncommitted=False
            )
            if ds.name is not None
            and ds.name.startswith(self.Constants.STEP_ARTIFACT_PREFIX)
            and not ds.name.endswith(self.Constants.LOCK_ARTIFACT_SUFFIX)
        )



import atexit
import json
import logging
import os.path
import tempfile
import time
import urllib
import urllib.parse
from pathlib import Path
from typing import Any, Dict, Optional, Union

from beaker import Beaker
from beaker import Dataset as BeakerDataset
from beaker import DatasetConflict, DatasetNotFound, Experiment, ExperimentNotFound

from tango.common.remote_utils import RemoteConstants
from tango.step import Step
from tango.step_info import StepInfo
from tango.version import VERSION

logger = logging.getLogger(__name__)


class Constants(RemoteConstants):
    ENTRYPOINT_DATASET_PREFIX = "tango-entrypoint-"
    BEAKER_TOKEN_SECRET_NAME: str = "BEAKER_TOKEN"
    GOOGLE_TOKEN_SECRET_NAME: str = "GOOGLE_TOKEN"
    DEFAULT_GOOGLE_CREDENTIALS_FILE: str = os.path.expanduser(
        os.path.join("~", ".config", "gcloud", "application_default_credentials.json")
    )
    ENTRYPOINT_DIR: str = "/tango/entrypoint"
    ENTRYPOINT_FILENAME: str = "entrypoint.sh"


def get_client(beaker_workspace: Optional[str] = None, **kwargs) -> Beaker:
    user_agent = f"tango v{VERSION}"
    if beaker_workspace is not None:
        return Beaker.from_env(
            default_workspace=beaker_workspace,
            session=True,
            user_agent=user_agent,
            **kwargs,
        )
    else:
        return Beaker.from_env(session=True, user_agent=user_agent, **kwargs)


def dataset_url(beaker: Beaker, dataset: Optional[str] = None) -> str:
    # this just creates a string url.
    workspace_url = beaker.workspace.url()
    if dataset:
        return (
            workspace_url
            + "/datasets?"
            + urllib.parse.urlencode(
                {
                    "text": dataset,
                    "committed": "false",
                }
            )
        )
    return workspace_url


class BeakerStepLock:
    METADATA_FNAME = "metadata.json"

    def __init__(
        self,
        beaker: Beaker,
        step: Union[str, StepInfo, Step],
        current_beaker_experiment: Optional[Experiment] = None,
    ):
        self._beaker = beaker
        self._step_id = step if isinstance(step, str) else step.unique_id
        self._lock_dataset_name = RemoteConstants.step_lock_artifact_name(step)
        self._lock_dataset: Optional[BeakerDataset] = None
        self._current_beaker_experiment = current_beaker_experiment
        self.lock_dataset_url = dataset_url(beaker, self._lock_dataset_name)

    @property
    def metadata(self) -> Dict[str, Any]:
        return {
            "beaker_experiment": None
            if not self._current_beaker_experiment
            else self._current_beaker_experiment.id
        }

    def _last_metadata(self) -> Optional[Dict[str, Any]]:
        try:
            metadata_bytes = self._beaker.dataset.get_file(
                self._lock_dataset_name, self.METADATA_FNAME, quiet=True
            )
            metadata = json.loads(metadata_bytes)
            return metadata
        except (DatasetNotFound, FileNotFoundError):
            return None

    def _acquiring_job_is_done(self) -> bool:
        last_metadata = self._last_metadata()
        if last_metadata is None:
            return False

        last_experiment_id = last_metadata.get("beaker_experiment")
        if last_experiment_id is None:
            return False

        try:
            last_experiment = self._beaker.experiment.get(last_experiment_id)
            if (
                self._current_beaker_experiment is not None
                and self._current_beaker_experiment.id == last_experiment_id
            ):
                # This means a previous job for this experiment was preempted and
                # it didn't clean up after itself.
                return True
            else:
                job = self._beaker.experiment.latest_job(last_experiment)
                return False if job is None else job.is_done
        except ExperimentNotFound:
            # Experiment must have been deleted.
            return True
        except ValueError:
            return False

    def acquire(self, timeout=None, poll_interval: float = 2.0, log_interval: float = 30.0) -> None:
        if self._lock_dataset is not None:
            return
        start = time.monotonic()
        last_logged = None
        while timeout is None or (time.monotonic() - start < timeout):
            try:
                self._lock_dataset = self._beaker.dataset.create(
                    self._lock_dataset_name, commit=False
                )

                atexit.register(self.release)

                # Write metadata.
                with tempfile.TemporaryDirectory() as tmp_dir_name:
                    tmp_dir = Path(tmp_dir_name)
                    metadata_path = tmp_dir / self.METADATA_FNAME
                    with open(metadata_path, "w") as f:
                        json.dump(self.metadata, f)
                    self._beaker.dataset.sync(self._lock_dataset, metadata_path, quiet=True)
            except DatasetConflict:
                # Check if existing lock was created from a Beaker experiment.
                # If it was, and the experiment is no-longer running, we can safely
                # delete it.
                if self._acquiring_job_is_done():
                    self._beaker.dataset.delete(self._lock_dataset_name)
                    continue

                now = time.monotonic()
                if last_logged is None or now - last_logged >= log_interval:
                    logger.warning(
                        "Waiting to acquire lock dataset for step '%s':\n\n%s\n\n"
                        "This probably means the step is being run elsewhere, but if you're sure it isn't "
                        "you can just delete the lock dataset.",
                        self._step_id,
                        self.lock_dataset_url,
                    )
                    last_logged = now
                time.sleep(poll_interval)
                continue
            else:
                break
        else:
            raise TimeoutError(
                f"Timeout error occurred while waiting to acquire dataset lock for step '{self._step_id}':\n\n"
                f"{self.lock_dataset_url}\n\n"
                f"This probably means the step is being run elsewhere, but if you're sure it isn't you can "
                f"just delete the lock dataset."
            )

    def release(self):
        if self._lock_dataset is not None:
            try:
                self._beaker.dataset.delete(self._lock_dataset)
            except DatasetNotFound:
                # Dataset must have been manually deleted.
                pass
            self._lock_dataset = None
            atexit.unregister(self.release)

    def __del__(self):
        self.release()



import json
import random
from pathlib import Path
from typing import Dict, Generator, Iterable, List, Optional, TypeVar, Union, cast
from urllib.parse import ParseResult

import petname
from google.auth.credentials import Credentials
from google.cloud import datastore

from tango.common.util import utc_now_datetime
from tango.integrations.gs.common import (
    Constants,
    GCSStepLock,
    get_client,
    get_credentials,
)
from tango.integrations.gs.step_cache import GSStepCache
from tango.step import Step
from tango.step_info import StepInfo, StepState
from tango.workspace import Run, RunInfo, RunSort, StepInfoSort, Workspace
from tango.workspaces.remote_workspace import RemoteWorkspace

T = TypeVar("T")


@Workspace.register("gs")
class GSWorkspace(RemoteWorkspace):
    """
    This is a :class:`~tango.workspace.Workspace` that stores step artifacts on Google Cloud Storage.

    .. tip::
        Registered as a :class:`~tango.workspace.Workspace` under the name "gs".

    :param workspace: The name or ID of the Google Cloud bucket to use.
    :param project: The Google project ID. This is required for the datastore. If not provided,
        it will be inferred from the Google cloud credentials.

    .. important::
        Credentials can be provided in the following ways:

        - Using the `credentials` keyword argument:
            - You can specify the path to the credentials json file.
            - You can specify the `google.oauth2.credentials.Credentials()` object.
            - You can specify the json string of credentials dict.

        - Using the default credentials: You can use your default google cloud credentials by running
          `gcloud auth application-default login`. If you are using `GSWorkspace` with
          :class:`~tango.integrations.beaker.BeakerExecutor`, you will need to set the environment variable
          `GOOGLE_TOKEN` to the credentials json file. The default location is usually
          `~/.config/gcloud/application_default_credentials.json`.

    """

    Constants = Constants
    NUM_CONCURRENT_WORKERS = 32

    def __init__(
        self,
        workspace: str,
        project: Optional[str] = None,
        credentials: Optional[Union[str, Credentials]] = None,
    ):
        self.client = get_client(bucket_name=workspace, credentials=credentials, project=project)

        self.client.NUM_CONCURRENT_WORKERS = self.NUM_CONCURRENT_WORKERS
        self._cache = GSStepCache(workspace, client=self.client)
        self._locks: Dict[Step, GCSStepLock] = {}

        super().__init__()

        credentials = get_credentials()
        project = project or credentials.quota_project_id
        self._ds = datastore.Client(namespace=workspace, project=project, credentials=credentials)

    @property
    def cache(self):
        return self._cache

    @property
    def locks(self):
        return self._locks

    @property
    def steps_dir_name(self):
        return "gs_workspace"

    @classmethod
    def from_parsed_url(cls, parsed_url: ParseResult) -> Workspace:
        workspace: str
        if parsed_url.netloc and parsed_url.path:
            # e.g. "gs://ai2/my-workspace"
            workspace = parsed_url.netloc + parsed_url.path
        elif parsed_url.netloc:
            # e.g. "gs://my-workspace"
            workspace = parsed_url.netloc
        else:
            raise ValueError(f"Bad URL for GS workspace '{parsed_url}'")
        return cls(workspace)

    @property
    def url(self) -> str:
        return self.client.url()

    def _remote_lock(self, step: Step) -> GCSStepLock:
        return GCSStepLock(self.client, step)

    def _step_location(self, step: Step) -> str:
        return self.client.url(self.Constants.step_artifact_name(step))

    def _save_run(
        self, steps: Dict[str, StepInfo], run_data: Dict[str, str], name: Optional[str] = None
    ) -> Run:
        if name is None:
            while True:
                name = petname.generate() + str(random.randint(0, 100))
                if not self._ds.get(self._ds.key("run", name)):
                    break
        else:
            if self._ds.get(self._ds.key("run", name)):
                raise ValueError(f"Run name '{name}' is already in use")

        run_entity = self._ds.entity(key=self._ds.key("run", name), exclude_from_indexes=("steps",))
        # Even though the run's name is part of the key, we add this as a
        # field so we can index on it and order asc/desc (indices on the key field don't allow ordering).
        run_entity["name"] = name
        run_entity["start_date"] = utc_now_datetime()
        run_entity["steps"] = json.dumps(run_data).encode()
        self._ds.put(run_entity)

        return Run(name=cast(str, name), steps=steps, start_date=run_entity["start_date"])

    def _get_run_from_entity(self, run_entity: datastore.Entity) -> Optional[Run]:
        try:
            steps_info_bytes = run_entity["steps"]
            steps_info = json.loads(steps_info_bytes)
        except KeyError:
            return None

        import concurrent.futures

        steps: Dict[str, StepInfo] = {}
        with concurrent.futures.ThreadPoolExecutor(
            max_workers=self.NUM_CONCURRENT_WORKERS,
            thread_name_prefix="GSWorkspace._get_run_from_dataset()-",
        ) as executor:
            step_info_futures = []
            for unique_id in steps_info.values():
                step_info_futures.append(executor.submit(self.step_info, unique_id))
            for future in concurrent.futures.as_completed(step_info_futures):
                step_info = future.result()
                assert step_info.step_name is not None
                steps[step_info.step_name] = step_info

        return Run(name=run_entity.key.name, start_date=run_entity["start_date"], steps=steps)

    def registered_runs(self) -> Dict[str, Run]:
        import concurrent.futures

        runs: Dict[str, Run] = {}

        with concurrent.futures.ThreadPoolExecutor(
            max_workers=self.NUM_CONCURRENT_WORKERS,
            thread_name_prefix="GSWorkspace.registered_runs()-",
        ) as executor:
            run_futures = []
            for run_entity in self._ds.query(kind="run").fetch():
                run_futures.append(executor.submit(self._get_run_from_entity, run_entity))
            for future in concurrent.futures.as_completed(run_futures):
                run = future.result()
                if run is not None:
                    runs[run.name] = run

        return runs

    def search_registered_runs(
        self,
        *,
        sort_by: Optional[RunSort] = None,
        sort_descending: bool = True,
        match: Optional[str] = None,
        start: int = 0,
        stop: Optional[int] = None,
    ) -> List[RunInfo]:
        run_entities = self._fetch_run_entities(
            sort_by=sort_by, sort_descending=sort_descending, match=match, start=start, stop=stop
        )
        return [
            RunInfo(name=e.key.name, start_date=e["start_date"], steps=json.loads(e["steps"]))
            for e in run_entities
        ]

    def num_registered_runs(self, *, match: Optional[str] = None) -> int:
        count = 0
        for _ in self._fetch_run_entities(match=match):
            count += 1
        return count

    def _fetch_run_entities(
        self,
        *,
        sort_by: Optional[RunSort] = None,
        sort_descending: bool = True,
        match: Optional[str] = None,
        start: int = 0,
        stop: Optional[int] = None,
    ) -> Generator[datastore.Entity, None, None]:
        from itertools import islice

        # Note: we can't query or order by multiple fields without a suitable
        # composite index. So in that case we have to apply remaining filters
        # or slice and order locally. We'll default to using 'match' in the query.
        # But if 'match' is null we can sort with the query.
        sort_locally = bool(match)

        sort_field: Optional[str] = None
        if sort_by == RunSort.START_DATE:
            sort_field = "start_date"
        elif sort_by == RunSort.NAME:
            sort_field = "name"
        elif sort_by is not None:
            raise NotImplementedError(sort_by)

        order: List[str] = []
        if sort_field is not None and not sort_locally:
            order = [sort_field if not sort_descending else f"-{sort_field}"]

        query = self._ds.query(kind="run", order=order)
        if match:
            # HACK: Datastore has no direct string matching functionality,
            # but this comparison is equivalent to checking if 'name' starts with 'match'.
            query.add_filter("name", ">=", match)
            query.add_filter("name", "<=", match[:-1] + chr(ord(match[-1]) + 1))

        entity_iter: Iterable[datastore.Entity] = query.fetch(
            offset=0 if sort_locally else start,
            limit=None if (stop is None or sort_locally) else stop - start,
        )

        if sort_field is not None and sort_locally:
            entity_iter = sorted(
                entity_iter, key=lambda entity: entity[sort_field], reverse=sort_descending
            )

        if sort_locally:
            entity_iter = islice(entity_iter, start, stop)

        for entity in entity_iter:
            yield entity

    def search_step_info(
        self,
        *,
        sort_by: Optional[StepInfoSort] = None,
        sort_descending: bool = True,
        match: Optional[str] = None,
        state: Optional[StepState] = None,
        start: int = 0,
        stop: Optional[int] = None,
    ) -> List[StepInfo]:
        step_info_entities = self._fetch_step_info_entities(
            sort_by=sort_by,
            sort_descending=sort_descending,
            match=match,
            state=state,
            start=start,
            stop=stop,
        )
        return [
            StepInfo.from_json_dict(json.loads(e["step_info_dict"])) for e in step_info_entities
        ]

    def num_steps(self, *, match: Optional[str] = None, state: Optional[StepState] = None) -> int:
        count = 0
        for _ in self._fetch_step_info_entities(match=match, state=state):
            count += 1
        return count

    def _fetch_step_info_entities(
        self,
        *,
        sort_by: Optional[StepInfoSort] = None,
        sort_descending: bool = True,
        match: Optional[str] = None,
        state: Optional[StepState] = None,
        start: int = 0,
        stop: Optional[int] = None,
    ) -> Generator[datastore.Entity, None, None]:
        from itertools import islice

        # Note: we can't query or order by multiple fields without a suitable
        # composite index. So in that case we have to apply remaining filters
        # or slice and order locally. We'll default to using 'match' in the query.
        # But if 'match' is null, we'll use 'state' to filter in the query.
        # If 'state' is also null, we can sort with the query.
        sort_locally = sort_by is not None and (match is not None or state is not None)
        filter_locally = state is not None and match is not None
        slice_locally = sort_locally or filter_locally

        sort_field: Optional[str] = None
        if sort_by == StepInfoSort.START_TIME:
            sort_field = "start_time"
        elif sort_by == StepInfoSort.UNIQUE_ID:
            sort_field = "step_id"
        elif sort_by is not None:
            raise NotImplementedError(sort_by)

        order: List[str] = []
        if sort_field is not None and not sort_locally:
            order = [sort_field if not sort_descending else f"-{sort_field}"]

        query = self._ds.query(kind="stepinfo", order=order)

        if match is not None:
            # HACK: Datastore has no direct string matching functionality,
            # but this comparison is equivalent to checking if 'step_id' starts with 'match'.
            query.add_filter("step_id", ">=", match)
            query.add_filter("step_id", "<=", match[:-1] + chr(ord(match[-1]) + 1))
        elif state is not None and not filter_locally:
            query.add_filter("state", "=", str(state.value))

        entity_iter: Iterable[datastore.Entity] = query.fetch(
            offset=0 if slice_locally else start,
            limit=None if (stop is None or slice_locally) else stop - start,
        )

        if state is not None and filter_locally:
            entity_iter = filter(lambda entity: entity["state"] == state, entity_iter)

        if sort_field is not None and sort_locally:
            entity_iter = sorted(
                entity_iter, key=lambda entity: entity[sort_field], reverse=sort_descending
            )

        if slice_locally:
            entity_iter = islice(entity_iter, start, stop)

        for entity in entity_iter:
            yield entity

    def registered_run(self, name: str) -> Run:
        err_msg = f"Run '{name}' not found in workspace"

        run_entity = self._ds.get(key=self._ds.key("run", name))
        if not run_entity:
            raise KeyError(err_msg)

        run = self._get_run_from_entity(run_entity)
        if run is None:
            raise KeyError(err_msg)
        else:
            return run

    def step_info(self, step_or_unique_id: Union[Step, str]) -> StepInfo:
        unique_id = (
            step_or_unique_id if isinstance(step_or_unique_id, str) else step_or_unique_id.unique_id
        )
        step_info_entity = self._ds.get(key=self._ds.key("stepinfo", unique_id))
        if step_info_entity is not None:
            step_info_bytes = step_info_entity["step_info_dict"]
            step_info = StepInfo.from_json_dict(json.loads(step_info_bytes))
            return step_info
        else:
            if not isinstance(step_or_unique_id, Step):
                raise KeyError(step_or_unique_id)
            step_info = StepInfo.new_from_step(step_or_unique_id)
            self._update_step_info(step_info)
            return step_info

    def _update_step_info(self, step_info: StepInfo):
        step_info_entity = self._ds.entity(
            key=self._ds.key("stepinfo", step_info.unique_id),
            exclude_from_indexes=("step_info_dict",),
        )

        # Even though the step's unique ID is part of the key, we add this as a
        # field so we can index on it and order asc/desc (indices on the key field don't allow ordering).
        step_info_entity["step_id"] = step_info.unique_id
        step_info_entity["step_name"] = step_info.step_name
        step_info_entity["start_time"] = step_info.start_time
        step_info_entity["end_time"] = step_info.end_time
        step_info_entity["state"] = str(step_info.state.value)
        step_info_entity["updated"] = utc_now_datetime()
        step_info_entity["step_info_dict"] = json.dumps(step_info.to_json_dict()).encode()

        self._ds.put(step_info_entity)

    def _save_run_log(self, name: str, log_file: Path):
        """
        The logs are stored in the bucket. The Run object details are stored in
        the remote database.
        """
        run_dataset = self.Constants.run_artifact_name(name)
        self.client.upload(run_dataset, log_file)



"""
A Python library for choreographing your machine learning research.
"""

__all__ = [
    "Format",
    "DillFormat",
    "DillFormatIterator",
    "Executor",
    "JsonFormat",
    "JsonFormatIterator",
    "SqliteDictFormat",
    "Step",
    "step",
    "StepInfo",
    "StepInfoSort",
    "StepState",
    "StepResources",
    "StepCache",
    "StepGraph",
    "Run",
    "RunInfo",
    "RunSort",
    "Workspace",
]

from .executor import Executor
from .format import (
    DillFormat,
    DillFormatIterator,
    Format,
    JsonFormat,
    JsonFormatIterator,
    SqliteDictFormat,
)
from .step import Step, StepResources, step
from .step_cache import StepCache
from .step_graph import StepGraph
from .step_info import StepInfo, StepState
from .workspace import Run, RunInfo, RunSort, StepInfoSort, Workspace



import logging
import os
import shutil
import tempfile
from abc import abstractmethod
from pathlib import Path
from typing import Any, Union

from tango.common.aliases import PathOrStr
from tango.common.exceptions import TangoError
from tango.common.file_lock import FileLock
from tango.common.params import Params
from tango.common.remote_utils import RemoteConstants
from tango.step import Step
from tango.step_cache import CacheMetadata
from tango.step_caches.local_step_cache import LocalStepCache
from tango.step_info import StepInfo

logger = logging.getLogger(__name__)


class RemoteNotFoundError(TangoError):
    """
    Classes inheriting from the RemoteStepCache should raise this if a step result object is not found.
    """


# This class inherits from `LocalStepCache` to benefit from its in-memory "weak cache" and "strong cache",
# but it handles saving artifacts to disk a little differently.
class RemoteStepCache(LocalStepCache):
    """
    This is a :class:`~tango.step_cache.StepCache` that's used by :class:`RemoteWorkspace`.
    It stores the results of steps on some RemoteWorkspace.

    It also keeps a limited in-memory cache as well as a local backup on disk, so fetching a
    step's resulting subsequent times should be fast.

    .. tip::
        All remote step caches inherit from this.
    """

    Constants = RemoteConstants

    def __init__(self, local_dir: Path):
        super().__init__(local_dir)

    @abstractmethod
    def _step_result_remote(self, step: Union[Step, StepInfo]):
        raise NotImplementedError()

    @abstractmethod
    def _upload_step_remote(self, step: Step, objects_dir: Path):
        raise NotImplementedError()

    @abstractmethod
    def _download_step_remote(self, step_result, target_dir: PathOrStr) -> None:
        raise NotImplementedError()

    @abstractmethod
    def __len__(self):
        raise NotImplementedError()

    def _acquire_step_lock_file(self, step: Union[Step, StepInfo], read_only_ok: bool = False):
        return FileLock(
            self.step_dir(step).with_suffix(".lock"), read_only_ok=read_only_ok
        ).acquire_with_updates(desc=f"acquiring step cache lock for '{step.unique_id}'")

    def __contains__(self, step: Any) -> bool:
        if isinstance(step, (Step, StepInfo)):
            cacheable = step.cache_results if isinstance(step, Step) else step.cacheable
            if not cacheable:
                return False

            key = step.unique_id

            # First check if we have a copy in memory.
            if key in self.strong_cache:
                return True
            if key in self.weak_cache:
                return True

            # Then check if we have a copy on disk in our cache directory.
            with self._acquire_step_lock_file(step, read_only_ok=True):
                if self.step_dir(step).is_dir():
                    return True

            # If not, check the remote location.
            return self._step_result_remote(step) is not None
        else:
            return False

    def __getitem__(self, step: Union[Step, StepInfo]) -> Any:
        key = step.unique_id
        step_result = self._step_result_remote(step)
        if step_result is None:
            raise KeyError(step)

        # Try getting the result from our in-memory caches first.
        result = self._get_from_cache(key)
        if result is not None:
            return result

        def load_and_return():
            metadata = CacheMetadata.from_params(Params.from_file(self._metadata_path(step)))
            result = metadata.format.read(self.step_dir(step) / self.Constants.STEP_RESULT_DIR)
            self._add_to_cache(key, result)
            return result

        # Next check our local on-disk cache.
        with self._acquire_step_lock_file(step, read_only_ok=True):
            if self.step_dir(step).is_dir():
                return load_and_return()

        # Finally, check the remote location for the corresponding dataset.
        with self._acquire_step_lock_file(step):
            # Make sure the step wasn't cached since the last time we checked (above).
            if self.step_dir(step).is_dir():
                return load_and_return()

            # We'll download the dataset to a temporary directory first, in case something goes wrong.
            temp_dir = tempfile.mkdtemp(dir=self.dir, prefix=key)
            try:
                self._download_step_remote(step_result, target_dir=temp_dir)
                # Download and extraction was successful, rename temp directory to final step result directory.
                os.replace(temp_dir, self.step_dir(step))
            except RemoteNotFoundError:
                raise KeyError(step)
            finally:
                shutil.rmtree(temp_dir, ignore_errors=True)

            return load_and_return()

    def __setitem__(self, step: Step, value: Any) -> None:
        if not step.cache_results:
            logger.warning("Tried to cache step %s despite being marked as uncacheable.", step.name)
            return

        with self._acquire_step_lock_file(step):
            # We'll write the step's results to temporary directory first, and try to upload to
            # remote workspace from there in case anything goes wrong.
            temp_dir = Path(tempfile.mkdtemp(dir=self.dir, prefix=step.unique_id))
            (temp_dir / self.Constants.STEP_RESULT_DIR).mkdir()
            try:
                step.format.write(value, temp_dir / self.Constants.STEP_RESULT_DIR)
                metadata = CacheMetadata(step=step.unique_id, format=step.format)
                metadata.to_params().to_file(temp_dir / self.METADATA_FILE_NAME)
                # Create the dataset and upload serialized result to it.
                self._upload_step_remote(step, temp_dir)
                # Upload successful, rename temp directory to the final step result directory.
                if self.step_dir(step).is_dir():
                    shutil.rmtree(self.step_dir(step), ignore_errors=True)
                os.replace(temp_dir, self.step_dir(step))
            finally:
                shutil.rmtree(temp_dir, ignore_errors=True)

        # Finally, add to in-memory caches.
        self._add_to_cache(step.unique_id, value)



import logging
import tempfile
import warnings
from abc import abstractmethod
from contextlib import contextmanager
from pathlib import Path
from typing import Dict, Generator, Iterable, Iterator, Optional, Tuple, TypeVar, Union
from urllib.parse import ParseResult

from tango.common.exceptions import StepStateError
from tango.common.logging import file_handler
from tango.common.remote_utils import RemoteConstants
from tango.common.util import exception_to_string, tango_cache_dir, utc_now_datetime
from tango.step import Step
from tango.step_caches.remote_step_cache import RemoteStepCache
from tango.step_info import StepInfo, StepState
from tango.workspace import Run, Workspace

T = TypeVar("T")

logger = logging.getLogger(__name__)


class RemoteWorkspace(Workspace):
    """
    This is a :class:`~tango.workspace.Workspace` that stores step artifacts on some remote storage location.

    .. tip::
        All remote workspaces inherit from this.
    """

    Constants = RemoteConstants
    NUM_CONCURRENT_WORKERS: int = 9

    @property
    @abstractmethod
    def cache(self) -> RemoteStepCache:
        raise NotImplementedError()

    @property
    @abstractmethod
    def steps_dir_name(self) -> str:
        raise NotImplementedError()

    @property
    @abstractmethod
    def locks(self) -> Dict:
        raise NotImplementedError()

    @property
    def steps_dir(self) -> Path:
        return tango_cache_dir() / self.steps_dir_name

    @property
    @abstractmethod
    def url(self) -> str:
        raise NotImplementedError()

    @classmethod
    @abstractmethod
    def from_parsed_url(cls, parsed_url: ParseResult) -> Workspace:
        raise NotImplementedError()

    @property
    def step_cache(self) -> RemoteStepCache:
        return self.cache

    def step_dir(self, step_or_unique_id: Union[Step, str]) -> Path:
        unique_id = (
            step_or_unique_id if isinstance(step_or_unique_id, str) else step_or_unique_id.unique_id
        )
        path = self.steps_dir / unique_id
        path.mkdir(parents=True, exist_ok=True)
        return path

    def work_dir(self, step: Step) -> Path:
        path = self.step_dir(step) / "work"
        path.mkdir(parents=True, exist_ok=True)
        return path

    def step_info(self, step_or_unique_id: Union[Step, str]) -> StepInfo:
        raise NotImplementedError()

    @abstractmethod
    def _remote_lock(self, step: Step):
        raise NotImplementedError()

    @abstractmethod
    def _step_location(self, step: Step) -> str:
        raise NotImplementedError()

    def step_starting(self, step: Step) -> None:
        # We don't do anything with uncacheable steps.
        if not step.cache_results:
            return

        # Get local file lock + remote dataset lock.
        lock = self._remote_lock(step)
        lock.acquire()
        self.locks[step] = lock

        step_info = self.step_info(step)
        if step_info.state == StepState.RUNNING:
            # Since we've acquired the step lock we know this step can't be running
            # elsewhere. But the step state can still say its running if the last
            warnings.warn(
                f"Step info for step '{step.unique_id}' is invalid - says step is running "
                "although it shouldn't be. Ignoring and overwriting step start time.",
                UserWarning,
            )
        elif step_info.state not in {StepState.INCOMPLETE, StepState.FAILED, StepState.UNCACHEABLE}:
            self.locks.pop(step).release()
            raise StepStateError(
                step,
                step_info.state,
                context=f"If you are certain the step is not running somewhere else, delete the step "
                f"datasets at {self._step_location(step)}",
            )

        if step_info.state == StepState.FAILED:
            # Refresh the environment metadata since it might be out-of-date now.
            step_info.refresh()

        # Update StepInfo to mark as running.
        try:
            step_info.start_time = utc_now_datetime()
            step_info.end_time = None
            step_info.error = None
            step_info.result_location = None
            self._update_step_info(step_info)
        except:  # noqa: E722
            self.locks.pop(step).release()
            raise

    def step_finished(self, step: Step, result: T) -> T:
        # We don't do anything with uncacheable steps.
        if not step.cache_results:
            return result

        step_info = self.step_info(step)
        if step_info.state != StepState.RUNNING:
            raise StepStateError(step, step_info.state)

        # Update step info and save step execution metadata.
        # This needs to be done *before* adding the result to the cache, since adding
        # the result to the cache will commit the step dataset, making it immutable.
        step_info.end_time = utc_now_datetime()
        step_info.result_location = self._step_location(step)
        self._update_step_info(step_info)

        self.cache[step] = result
        if hasattr(result, "__next__"):
            assert isinstance(result, Iterator)
            # Caching the iterator will consume it, so we write it to the cache and then read from the cache
            # for the return value.
            result = self.cache[step]

        self.locks.pop(step).release()

        return result

    def step_failed(self, step: Step, e: BaseException) -> None:
        # We don't do anything with uncacheable steps.
        if not step.cache_results:
            return

        try:
            step_info = self.step_info(step)
            if step_info.state != StepState.RUNNING:
                raise StepStateError(step, step_info.state)
            step_info.end_time = utc_now_datetime()
            step_info.error = exception_to_string(e)
            self._update_step_info(step_info)
        finally:
            self.locks.pop(step).release()

    def _get_run_step_info(self, targets: Iterable[Step]) -> Tuple[Dict, Dict]:
        import concurrent.futures

        all_steps = set(targets)
        for step in targets:
            all_steps |= step.recursive_dependencies

        steps: Dict[str, StepInfo] = {}
        run_data: Dict[str, str] = {}

        # Collect step info.
        with concurrent.futures.ThreadPoolExecutor(
            thread_name_prefix="RemoteWorkspace._get_run_step_info()-"
        ) as executor:
            step_info_futures = []
            for step in all_steps:
                if step.name is None:
                    continue
                step_info_futures.append(executor.submit(self.step_info, step))
            for future in concurrent.futures.as_completed(step_info_futures):
                step_info = future.result()
                assert step_info.step_name is not None
                steps[step_info.step_name] = step_info
                run_data[step_info.step_name] = step_info.unique_id

        return steps, run_data

    @abstractmethod
    def _save_run(
        self, steps: Dict[str, StepInfo], run_data: Dict[str, str], name: Optional[str] = None
    ) -> Run:
        raise NotImplementedError()

    def register_run(self, targets: Iterable[Step], name: Optional[str] = None) -> Run:
        steps, run_data = self._get_run_step_info(targets)
        run = self._save_run(steps, run_data, name)
        return run

    @abstractmethod
    def _save_run_log(self, name: str, log_file: Path):
        raise NotImplementedError()

    @contextmanager
    def capture_logs_for_run(self, name: str) -> Generator[None, None, None]:
        with tempfile.TemporaryDirectory() as tmp_dir_name:
            log_file = Path(tmp_dir_name) / "out.log"
            try:
                with file_handler(log_file):
                    yield None
            finally:
                self._save_run_log(name, log_file)

    @abstractmethod
    def _update_step_info(self, step_info: StepInfo):
        raise NotImplementedError()



import logging
from pathlib import Path
from typing import Optional, Union

from beaker import Beaker
from beaker import Dataset as BeakerDataset
from beaker import DatasetConflict, DatasetNotFound, DatasetWriteError

from tango import Step
from tango.common import PathOrStr
from tango.common.exceptions import ConfigurationError
from tango.common.util import make_safe_filename, tango_cache_dir
from tango.integrations.beaker.common import Constants, get_client
from tango.step_cache import StepCache
from tango.step_caches.remote_step_cache import RemoteNotFoundError, RemoteStepCache
from tango.step_info import StepInfo

logger = logging.getLogger(__name__)


@StepCache.register("beaker")
class BeakerStepCache(RemoteStepCache):
    """
    This is a :class:`~tango.step_cache.StepCache` that's used by :class:`BeakerWorkspace`.
    It stores the results of steps on Beaker as datasets.

    It also keeps a limited in-memory cache as well as a local backup on disk, so fetching a
    step's resulting subsequent times should be fast.

    .. tip::
        Registered as a :class:`~tango.step_cache.StepCache` under the name "beaker".

    :param workspace: The name or ID of the Beaker workspace to use.
    :param beaker: The Beaker client to use.
    """

    Constants = Constants

    def __init__(self, beaker_workspace: Optional[str] = None, beaker: Optional[Beaker] = None):
        self.beaker: Beaker
        if beaker is not None:
            self.beaker = beaker
            if beaker_workspace is not None:
                self.beaker.config.default_workspace = beaker_workspace
                self.beaker.workspace.ensure(beaker_workspace)
        else:
            self.beaker = get_client(beaker_workspace=beaker_workspace)
        if self.beaker.config.default_workspace is None:
            raise ConfigurationError("Beaker default workspace must be set")
        super().__init__(
            tango_cache_dir()
            / "beaker_cache"
            / make_safe_filename(self.beaker.config.default_workspace)
        )

    def _step_result_remote(self, step: Union[Step, StepInfo]) -> Optional[BeakerDataset]:
        """
        Returns a `BeakerDataset` object containing the details of the step.
        This only returns if the step has been finalized (committed).
        """
        try:
            dataset = self.beaker.dataset.get(self.Constants.step_artifact_name(step))
            return dataset if dataset.committed is not None else None
        except DatasetNotFound:
            return None

    def _upload_step_remote(self, step: Step, objects_dir: Path) -> BeakerDataset:
        """
        Uploads the step's output to remote location.
        """
        dataset_name = self.Constants.step_artifact_name(step)
        try:
            self.beaker.dataset.create(dataset_name, commit=False)
        except DatasetConflict:
            pass
        try:
            self.beaker.dataset.sync(dataset_name, objects_dir, quiet=True)
            self.beaker.dataset.commit(dataset_name)
        except DatasetWriteError:
            pass

        return self.beaker.dataset.get(dataset_name)

    def _download_step_remote(self, step_result, target_dir: PathOrStr) -> None:
        """
        Downloads the step's output from remote location.
        """
        try:
            self.beaker.dataset.fetch(step_result, target_dir, quiet=True)
        except DatasetNotFound:
            raise RemoteNotFoundError()

    def __len__(self):
        """
        Returns the number of committed step outputs present in the remote location.
        """
        # NOTE: lock datasets should not count here.
        return sum(
            1
            for ds in self.beaker.workspace.iter_datasets(
                match=self.Constants.STEP_ARTIFACT_PREFIX, uncommitted=False, results=False
            )
            if ds.name is not None
            and ds.name.startswith(self.Constants.STEP_ARTIFACT_PREFIX)
            and not ds.name.endswith(self.Constants.LOCK_ARTIFACT_SUFFIX)
        )



import logging
from typing import Any, Optional, Union

import wandb
from retry import retry
from wandb.errors import Error as WandbError

from tango.common.aliases import PathOrStr
from tango.common.util import make_safe_filename, tango_cache_dir
from tango.step import Step
from tango.step_cache import StepCache
from tango.step_caches.remote_step_cache import RemoteNotFoundError, RemoteStepCache
from tango.step_info import StepInfo

from .util import ArtifactKind, check_environment, is_missing_artifact_error

logger = logging.getLogger(__name__)


@StepCache.register("wandb")
class WandbStepCache(RemoteStepCache):
    """
    This is a :class:`~tango.step_cache.StepCache` that's used by :class:`WandbWorkspace`.
    It stores the results of steps on W&B as Artifacts.

    It also keeps a limited in-memory cache as well as a local backup on disk, so fetching a
    step's resulting subsequent times should be fast.

    :param project: The W&B project to use.
    :param entity: The W&B entity (user or organization account) to use.

    .. tip::
        Registered as :class:`~tango.step_cache.StepCache` under the name "wandb".
    """

    def __init__(self, project: str, entity: str):
        check_environment()
        super().__init__(
            tango_cache_dir()
            / "wandb_cache"
            / make_safe_filename(entity)
            / make_safe_filename(project)
        )
        self.project = project
        self.entity = entity

    @property
    def wandb_client(self) -> wandb.Api:
        return wandb.Api(overrides={"entity": self.entity, "project": self.project})

    @property
    def client(self):
        """
        To maintain compatibility
        """
        return self.wandb_client

    @property
    def wandb_project_url(self) -> str:
        """
        The URL of the W&B project this workspace uses.
        """
        app_url = self.wandb_client.client.app_url
        app_url = app_url.rstrip("/")
        return f"{app_url}/{self.entity}/{self.project}"

    def _step_artifact_name(self, step: Union[Step, StepInfo]) -> str:
        if isinstance(step, Step):
            return step.class_name
        else:
            return step.step_class_name

    def _step_result_remote(  # type: ignore
        self, step: Union[Step, StepInfo]
    ) -> Optional[wandb.apis.public.Artifact]:
        artifact_kind = (step.metadata or {}).get("artifact_kind", ArtifactKind.STEP_RESULT.value)
        try:
            return self.wandb_client.artifact(
                f"{self.entity}/{self.project}/{self._step_artifact_name(step)}:{step.unique_id}",
                type=artifact_kind,
            )
        except WandbError as exc:
            if is_missing_artifact_error(exc):
                return None
            else:
                raise

    def create_step_result_artifact(self, step: Step, objects_dir: Optional[PathOrStr] = None):
        self._upload_step_remote(step, objects_dir)

    def get_step_result_artifact(
        self, step: Union[Step, StepInfo]
    ) -> Optional[wandb.apis.public.Artifact]:
        artifact_kind = (step.metadata or {}).get("artifact_kind", ArtifactKind.STEP_RESULT.value)
        try:
            return self.wandb_client.artifact(
                f"{self.entity}/{self.project}/{self._step_artifact_name(step)}:{step.unique_id}",
                type=artifact_kind,
            )
        except WandbError as exc:
            if is_missing_artifact_error(exc):
                return None
            else:
                raise

    def _upload_step_remote(self, step: Step, objects_dir: Optional[PathOrStr] = None) -> Any:
        """
        Create an artifact for the result of a step.
        """
        artifact_kind = (step.metadata or {}).get("artifact_kind", ArtifactKind.STEP_RESULT.value)
        artifact = wandb.Artifact(self._step_artifact_name(step), type=artifact_kind)

        # Add files
        if objects_dir is not None:
            artifact.add_dir(str(objects_dir))

        # Log/persist the artifact to W&B.
        artifact.save()
        artifact.wait()

        # Add an alias for the step's unique ID.
        # Only after we've logged the artifact can we add an alias.
        artifact.aliases.append(step.unique_id)
        artifact.save()
        artifact.wait()

    def get_step_result_artifact_url(self, step: Union[Step, StepInfo]) -> str:
        artifact_kind = (step.metadata or {}).get("artifact_kind", ArtifactKind.STEP_RESULT.value)
        return (
            f"{self.wandb_project_url}/artifacts/{artifact_kind}"
            f"/{self._step_artifact_name(step)}/{step.unique_id}"
        )

    @retry(exceptions=(wandb.errors.CommError,), delay=10, backoff=2, max_delay=120)
    def use_step_result_artifact(self, step: Union[Step, StepInfo]) -> None:
        """
        "Use" the artifact corresponding to the result of a step.
        """
        if wandb.run is None:
            raise RuntimeError("This can only be called from within a W&B run")
        wandb.run.use_artifact(
            f"{self.entity}/{self.project}/{self._step_artifact_name(step)}:{step.unique_id}"
        )

    def _download_step_remote(self, step_result, target_dir: PathOrStr):
        try:
            step_result.download(root=target_dir, recursive=True)
        except (WandbError, ValueError):
            raise RemoteNotFoundError()

    def __len__(self) -> int:
        completed_cacheable_step_runs = self.wandb_client.runs(
            f"{self.entity}/{self.project}",
            filters={  # type: ignore
                "config.job_type": "step",
                "config.cacheable": True,
                "state": "finished",
            },
        )
        return len(list(completed_cacheable_step_runs))



import logging
from pathlib import Path
from typing import Any, Dict, List, Optional

from tango.common.dataset_dict import DatasetDictBase
from tango.common.registrable import Registrable
from tango.workspace import Workspace

from .data import DataLoader
from .exceptions import StopEarly
from .model import Model
from .train_config import TrainConfig
from .training_engine import TrainingEngine


class TrainCallback(Registrable):
    """
    A :class:`TrainCallback` is a :class:`~tango.common.Registrable` class
    that can be used within :class:`TorchTrainStep` to customize behavior in the training
    loop. You can set the training callbacks with the ``callbacks`` parameter to :class:`TorchTrainStep`.

    .. tip::
        All of the parameters to this base class will be automatically set within
        the training loop, so you shouldn't include them in your config for your callbacks.

    .. tip::
        You can access the model being trained through :attr:`self.model <model>`.

    .. important::
        The ``step`` argument to callback methods is the total/overall number of training steps
        so far, independent of the current epoch.

    .. seealso::
        See :class:`~tango.integrations.wandb.WandbTrainCallback` for an example
        implementation.

    :ivar Workspace workspace: The tango workspace being used.
    :ivar TrainConfig train_config: The training config.
    :ivar TrainingEngine training_engine: The engine used to train the model.
    :ivar tango.common.DatasetDictBase dataset_dict: The dataset dict containing train and
        optional validation splits.
    :ivar DataLoader train_dataloader: The dataloader used for the training split.
    :ivar DataLoader validation_dataloader: Optional dataloader used for the validation split.
    """

    def __init__(
        self,
        workspace: Workspace,
        train_config: TrainConfig,
        training_engine: TrainingEngine,
        dataset_dict: DatasetDictBase,
        train_dataloader: DataLoader,
        validation_dataloader: Optional[DataLoader] = None,
    ) -> None:
        self.workspace = workspace
        self.train_config = train_config
        self.training_engine = training_engine
        self.dataset_dict = dataset_dict
        self.train_dataloader = train_dataloader
        self.validation_dataloader = validation_dataloader
        self.logger = logging.getLogger(self.__class__.__name__)

    @property
    def step_id(self) -> str:
        """
        The unique ID of the current :class:`~tango.Step`.
        """
        return self.train_config.step_id

    @property
    def step_name(self) -> Optional[str]:
        """
        The name of the current :class:`~tango.Step`.
        """
        return self.train_config.step_name

    @property
    def work_dir(self) -> Path:
        """
        The working directory of the current train step.
        """
        return self.train_config.work_dir

    @property
    def is_local_main_process(self) -> bool:
        """
        This is ``True`` if the current worker is the main distributed worker of the current node, or if
        we are not using distributed training.
        """
        return self.train_config.is_local_main_process

    @property
    def model(self) -> Model:
        """
        The :class:`Model` being trained.
        """
        return self.training_engine.model

    def state_dict(self) -> Dict[str, Any]:
        """
        Return any state that needs to be kept after a restart.

        Some callbacks need to maintain state across restarts. This is the callback's opportunity to
        save it's state. It will be restored using :meth:`load_state_dict`.
        """
        return {}

    def load_state_dict(self, state_dict: Dict[str, Any]) -> None:
        """
        Load the state on a restart.

        Some callbacks need to maintain state across restarts. This is the callback's opportunity to
        restore it's state. It gets saved using :meth:`state_dict`.
        """
        pass

    def pre_train_loop(self) -> None:
        """
        Called right before the first batch is processed, or after a restart.
        """
        pass

    def post_train_loop(self, step: int, epoch: int) -> None:
        """
        Called after the training loop completes.

        This is the last method that is called, so any cleanup can be done in this method.
        """
        pass

    def pre_epoch(self, step: int, epoch: int) -> None:
        """
        Called right before the start of an epoch. Epochs start at 0.
        """
        pass

    def post_epoch(self, step: int, epoch: int) -> None:
        """
        Called after an epoch is completed. Epochs start at 0.
        """
        pass

    def pre_batch(self, step: int, epoch: int, batch: List[Dict[str, Any]]) -> None:
        """
        Called directly before processing a batch.

        .. note::
            A type of ``batch`` is a list because with gradient accumulation there will
            more than one "micro batch" in the batch.

        """
        pass

    def post_batch(
        self, step: int, epoch: int, batch_loss: float, batch_outputs: List[Dict[str, Any]]
    ) -> None:
        """
        Called directly after processing a batch, but before unscaling gradients,
        clipping gradients, and taking an optimizer step.

        .. note::
            The ``batch_loss`` here is the loss local to the current worker, not the
            overall (average) batch loss across distributed workers.

            If you need the average loss, use :meth:`log_batch()`.

        .. note::
            A type of ``batch_outputs`` is a list because with gradient accumulation there will
            more than one "micro batch" in the batch.

        """
        pass

    def log_batch(
        self, step: int, epoch: int, batch_loss: float, batch_outputs: List[Dict[str, Any]]
    ) -> None:
        """
        Called after the optimizer step. Here ``batch_loss`` is the average loss across
        all distributed workers.

        .. note::
            This callback method is not necessarily called on every step.
            The frequency depends on the value of the ``log_every`` parameter of
            :class:`TorchTrainStep`.

        .. note::
            A type of ``batch_outputs`` is a list because with gradient accumulation there will
            more than one "micro batch" in the batch.

        """
        pass

    def pre_val_batch(
        self, step: int, val_step: int, epoch: int, val_batch: Dict[str, Any]
    ) -> None:
        """
        Called right before a validation batch is processed.
        """
        pass

    def post_val_batch(
        self, step: int, val_step: int, epoch: int, val_batch_outputs: Dict[str, Any]
    ) -> None:
        """
        Called right after a validation batch is processed with the outputs of the batch.

        .. tip::
            This method can be used to modify ``val_batch_outputs`` in place, which is useful
            in scenarios like distributed training where you might need to aggregate metrics
            in a special way other than a simple average. If that's the case, make sure
            to set ``auto_aggregate_val_metric`` to ``False`` in :class:`TorchTrainStep`.

        """
        pass

    def post_val_loop(
        self, step: int, epoch: int, val_metric: float, best_val_metric: float
    ) -> None:
        """
        Called right after the validation loop finishes.
        """
        pass


@TrainCallback.register("torch::stop_early")
class StopEarlyCallback(TrainCallback):
    """
    A :class:`TrainCallback` for early stopping. Training is stopped early after
    ``patience`` steps without an improvement to the validation metric.

    .. tip::

        Registered as a :class:`TrainCallback` under the name "torch::stop_early".
    """

    def __init__(self, *args, patience: int = 10000, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.patience = patience
        self.best_step = 0
        self.best_val_metric: Optional[float] = None

    def post_val_loop(
        self, step: int, epoch: int, val_metric: float, best_val_metric: float
    ) -> None:
        # We can't rely on the best_val_metric parameter, because then we can't detect when the metric stays
        # the same for many steps.
        if self.best_val_metric is None or val_metric > self.best_val_metric:
            self.best_step = step
            self.best_val_metric = val_metric
        elif step > self.best_step + self.patience:
            raise StopEarly

    def state_dict(self) -> Dict[str, Any]:
        """
        Return any state that needs to be kept after a restart.
        """
        return {
            "patience": self.patience,
            "best_step": self.best_step,
            "best_val_metric": self.best_val_metric,
        }

    def load_state_dict(self, state_dict: Dict[str, Any]) -> None:
        """
        Load the state on a restart.
        """
        self.patience = state_dict["patience"]
        self.best_step = state_dict["best_step"]
        self.best_val_metric = state_dict["best_val_metric"]
